from typing import List


class ai_prompt:
    def __init__(self, objective: List[str]):
        self.objective = objective

    def create(
        self,
        function_name: str,
    ) -> str:
        return " ".join(self.objective).replace("{function_name}", function_name)



import random
import numpy as np
import math
from openai_commands.logger import logger


class Brush:
    def __init__(
        self,
        canvas,
        width=256,
        height=256,
    ):
        self.width = width
        self.height = height

        self.cursor = [
            canvas.width // 2,
            canvas.height // 2,
        ]

    def move(self, canvas):
        self.cursor = [
            min(
                canvas.width - self.width,
                max(self.width, int(self.cursor[0])),
            ),
            min(
                canvas.height - self.height,
                max(self.height, int(self.cursor[1])),
            ),
        ]

        return self


class RandomWalkBrush(Brush):
    def __init__(
        self,
        canvas,
        width=256,
        height=256,
    ):
        super().__init__(canvas, width, height)

        self.horizontal_move = 1
        self.vertical_move = 1

    def move(self, canvas):
        self.cursor = [
            random.uniform(
                -self.horizontal_move * self.width,
                self.horizontal_move * self.width,
            )
            + self.cursor[0],
            random.uniform(
                -self.horizontal_move * self.height,
                self.horizontal_move * self.height,
            )
            + self.cursor[1],
        ]

        return super().move(canvas)


class TilingBrush(Brush):
    def __init__(
        self,
        canvas,
        width=256,
        height=256,
    ):
        super().__init__(canvas, width, height)

        self.ring = 0
        self.delta_angle = 2 * np.pi
        self.angle = 2 * np.pi

        self.ring_size = 0.7

        self.verbose = canvas.verbose

    def move(self, canvas):
        self.angle += self.delta_angle

        if self.angle >= 2 * np.pi - self.delta_angle / 2:
            self.ring += 1
            self.delta_angle = 2 * math.asin(1 / 2 / self.ring)
            self.angle = 0

        if self.verbose:
            logger.info(
                "TilingBrush: ring:{} @ {:.02f} deg".format(
                    self.ring, self.angle / np.pi * 180
                )
            )

        self.cursor[0] = (
            canvas.width // 2
            + self.ring * self.ring_size * math.cos(self.angle) * self.width
        )
        self.cursor[1] = (
            canvas.height // 2
            + self.ring * self.ring_size * math.sin(self.angle) * self.height
        )

        return super().move(canvas)



from blue_objects import file, objects


class bash_prompt:
    def __init__(self, prompt: str):
        self.prompt = prompt

    def create(
        self,
        function_name: str,
        function_short_name: str,
    ) -> str:
        success, instructions = file.load_text(
            objects.path_of(f"{function_short_name}-description.txt"),
            log=True,
        )
        assert success

        return "\n".join(
            [
                (
                    line.replace(function_name, function_short_name)
                    if function_short_name
                    else line
                )
                for line in instructions + [self.prompt]
            ]
        )

    @staticmethod
    def pre_process(filename):
        success, content = file.load_yaml(filename)
        return (
            success
            if not success
            else file.save_text(
                file.add_extension(filename, "txt"),
                [
                    "to {}, run {}.".format(
                        thing[1][:-1] if thing[1].endswith(".") else thing[1],
                        thing[0],
                    )
                    .replace("\\n", "")
                    .replace("\\t", "")
                    .replace("\\", "")
                    for thing in content
                ],
                log=True,
            )
        )



from io import BytesIO
from PIL import Image
from tqdm import tqdm
import requests
import numpy as np

from blueness import module
from blue_options.host import is_jupyter
from blue_options.host import signature as host_signature
from blue_options.logger import crash_report
from blue_objects import file
from blue_objects.objects import signature as object_signature
from blue_objects.graphics import add_signature

from openai_commands import NAME, VERSION
from openai_commands.DALLE.brush import RandomWalkBrush, TilingBrush
from openai_commands.logger import logger

NAME = module.name(__file__, NAME)


class Canvas:
    def __init__(
        self,
        shape=(2048, 4096),
        verbose=False,
        debug_mode=False,
        dryrun=False,
        content=None,
        source="",
        brush_kind="tiling",
        brush_size=256,
    ):
        self.verbose = verbose
        self.debug_mode = debug_mode
        self.dryrun = dryrun
        self.is_jupyter = is_jupyter()
        self.source = source
        self.content = content

        if content is not None:
            shape = Canvas.shape(
                content,
                brush_kind,
                brush_size,
            )

        self.height, self.width = shape

        self.allocated = not dryrun

        self.image = Image.new(
            "RGB",
            (
                0 if dryrun else self.width,
                0 if dryrun else self.height,
            ),
            (0, 0, 0),
        )
        self.mask = Image.new(
            "L",
            (self.width, self.height),
            (0,),
        )

        logger.info(
            "DALL-E Canvas({}x{}): {}allocated".format(
                self.height,
                self.width,
                "" if self.allocated else "not ",
            )
        )

        self.crashed = False

    def add_signature(self, image):
        return Image.fromarray(
            add_signature(
                np.array(image),
                [" | ".join(object_signature())],
                [
                    "{} : {}".format(
                        self.content[0] if len(self.content) >= 1 else "",
                        self.source,
                    )
                ]
                + [
                    " | ".join([f"{NAME}-{VERSION}"] + host_signature()),
                ],
            )
        )

    def box(self):
        indices = np.nonzero(np.array(self.mask) == 255)

        left = indices[1].min()
        top = indices[0].min()
        right = indices[1].max()
        bottom = indices[0].max()

        return (left, top, right, bottom)

    def create_brush(
        self,
        brush_kind="tiling",
        brush_size=256,
    ):
        if brush_kind == "tiling":
            return TilingBrush(self, brush_size, brush_size)

        if brush_kind == "randomwalk":
            return RandomWalkBrush(self, brush_size, brush_size)

        raise ValueError(
            f"-DALL-E: Canvas: create_brush: {brush_kind}: kind not found."
        )

    def paint(self, brush, prompt):
        from openai import OpenAI

        client = OpenAI()

        box = (
            brush.cursor[0] - brush.width // 2,  # left,
            brush.cursor[1] - brush.height // 2,  # top,
            brush.cursor[0] + brush.width // 2,  # right,
            brush.cursor[1] + brush.height // 2,  # bottom,
        )

        image_ = self.image.crop(box)
        image_byte_stream = BytesIO()
        image_.save(image_byte_stream, format="PNG")
        image_byte_array = image_byte_stream.getvalue()

        mask_ = image_.copy().convert("RGBA")
        mask_.putalpha(self.mask.crop(box))
        mask_byte_stream = BytesIO()
        mask_.save(mask_byte_stream, format="PNG")
        mask_byte_array = mask_byte_stream.getvalue()

        if not self.dryrun:
            try:
                response = client.images.generate(
                    image=image_byte_array,
                    mask=mask_byte_array,
                    prompt=prompt,
                    n=1,
                    size=f"{brush.width}x{brush.height}",
                )

                image_url = response.data[0].url
                if self.verbose:
                    logger.info(f"Canvas.paint: received {image_url}")

                response = requests.get(image_url)
                image_data = response.content
                image__ = Image.open(BytesIO(image_data))
                if self.verbose:
                    logger.info(
                        f"Canvas.paint: downloaded {image__.size}, {image__.mode}"
                    )

                self.image.paste(image__, box)
            except:
                crash_report(f"{NAME}: paint({prompt})")
                self.crashed = True

        self.mask.paste(
            Image.new(
                "L",
                (brush.width, brush.height),
                (255,),
            ),
            box,
        )

        if self.is_jupyter and self.verbose and not self.crashed:
            from IPython.display import display, clear_output

            clear_output(wait=True)

            if self.debug_mode:
                image = Image.new("RGB", (3 * brush.width, brush.height))
                image.paste(image_, (0, 0))
                image.paste(mask_, (brush.width, 0))
                image.paste(image__, (2 * brush.width, 0))
                display(self.add_signature(image))

                image = Image.new("RGB", (2 * self.width, self.height))
                image.paste(self.mask, (0, 0))
                image.paste(self.image, (self.width, 0))
                display(self.add_signature(image))
            else:
                display(self.add_signature(self.image.crop(self.box())))

        return self

    def render_text(
        self,
        brush,
        content,
        filename,
    ):
        content = [line for line in content if line]

        logger.info(f"Canvas.render_text: {len(content)} line(s).")

        for index in tqdm(range(len(content))):
            if self.verbose:
                logger.info(f"DALL-E: {content[index]}")

            self.paint(brush, content[index])

            brush.move(self)

            if self.verbose:
                self.save(filename)

        self.save(filename)

        return self

    def save(self, filename):
        box = self.box()

        mask = self.mask.crop(box)
        image = self.image.crop(box)

        image = self.add_signature(image)

        image.save(filename)
        mask.save(file.add_suffix(filename, "mask"))

        if self.debug_mode and self.is_jupyter:
            from IPython.display import display, clear_output

            clear_output(wait=True)
            display(image)

        logger.info(f"Canvas -> {filename}")

        file.save_json(
            file.add_extension(filename, "json"),
            {
                "content": self.content,
                "source": self.source,
                "height": self.height,
                "width": self.width,
                "generator": f"DALL-E-{VERSION}",
            },
        )

    @staticmethod
    def shape(
        content,
        brush_kind="tiling",
        brush_size=256,
        margin=0.0,
    ):
        canvas = Canvas(
            (25000, 25000),
            dryrun=True,
        )

        brush = canvas.create_brush(
            brush_kind,
            brush_size,
        )

        content = [line for line in content if line]

        for index in tqdm(range(len(content))):
            canvas.paint(brush, content[index])
            brush.move(canvas)

        left, top, right, bottom = canvas.box()

        plus_margin = 1 + 2 * margin

        height = int(
            2
            * max(
                canvas.height // 2 - top,
                bottom - canvas.height // 2,
            )
            * plus_margin
        )
        width = int(
            2
            * max(
                canvas.width // 2 - left,
                right - canvas.height // 2,
            )
            * plus_margin
        )

        logger.info(
            f"Canvas.shape: {len(content)} line(s) @ {brush_kind} x {brush_size}: {height}x{width}"
        )

        return height, width



import json
from openai import OpenAI
from typing import Tuple, Any
from IPython.display import Image, display

from blueness import module
from blue_options import string, host
from blue_options.host import is_jupyter
from blue_objects import file, graphics, objects

from openai_commands import NAME, VERSION
from openai_commands import env
from openai_commands.logger import logger


NAME = module.name(__file__, NAME)


class OpenAIImageGenerator:
    def __init__(
        self,
        model="dall-e-3",
        verbose: bool = False,
    ):
        self.client = OpenAI(api_key=env.OPENAI_API_KEY)
        self.verbose = verbose
        self.model = model
        logger.info(
            "{}[{}]".format(
                self.__class__.__name__,
                self.model,
            )
        )

    def generate(
        self,
        prompt: str,
        filename: str = "",
        sign: bool = True,
    ) -> Tuple[bool, Any]:
        logger.info(
            "{}.generate: {}".format(
                self.__class__.__name__,
                prompt,
            )
        )

        augmented_prompt = self.augment_prompt(prompt)

        response = self.client.images.generate(
            model=self.model,
            prompt=augmented_prompt,
            size="1024x1024",
            quality="standard",
            n=1,
        )

        logger.info(json.dumps(response.dict(), indent=4))

        success = file.download(response.data[0].url, filename) if filename else True

        if success and sign:
            success, image = file.load_image(filename)

        if success and sign:
            image = graphics.add_signature(
                image,
                [
                    " | ".join(host.signature()),
                    " | ".join(objects.signature(file.name(filename))),
                ],
                [
                    prompt,
                    str(response.data[0].revised_prompt),
                    " | ".join(
                        [
                            f"{NAME}-{VERSION}",
                            self.model,
                            string.pretty_shape_of_matrix(image),
                        ]
                    ),
                ],
            )
            success = file.save_image(filename, image)

        if success and self.verbose and is_jupyter():
            display(Image(filename=filename))

        return (
            success,
            response,
        )

    def augment_prompt(self, prompt: str) -> str:
        # https://community.openai.com/t/api-image-generation-in-dall-e-3-changes-my-original-prompt-without-my-permission/476355
        prompt = " ".join(
            [
                "THIS IS A RAW EXECUTION. ABSOLUTELY DO NOT REVISE THIS PROMPT:",
                prompt,
            ]
        )

        if self.verbose:
            logger.info(
                "{}.augment_prompt: {}".format(
                    self.__class__.__name__,
                    prompt,
                )
            )

        return prompt



from typing import Tuple, Dict, Any
from openai_commands.completion.functions.generic import ai_function


class ai_function_bash(ai_function):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.language = "bash"



import json
from enum import Enum, auto
from typing import List
from tqdm import tqdm
from blueness import module
from openai import OpenAI

from blue_objects.env import ABCLI_PUBLIC_PREFIX
from blue_options.options import Options
from blue_options import string
from blue_objects import file, objects

from openai_commands import env, NAME
from openai_commands.logger import logger

NAME = module.name(__file__, NAME)


class Detail(Enum):
    AUTO = auto()
    LOW = auto()
    HIGH = auto()


def complete_object(
    prompt: str,
    object_name: str,
    options: Options,
    max_count: int = 5,
    detail: Detail = Detail.AUTO,
    verbose: bool = False,
):
    logger.info(
        '{}.complete_object[{}]: "{}" @ {} < {} : {}'.format(
            NAME,
            detail,
            prompt,
            options.to_str(),
            max_count,
            object_name,
        )
    )

    options["inference"] = False
    list_of_images = objects.list_of_files(object_name, cloud=True)
    for keyword in options:
        list_of_images = [
            filename
            for filename in list_of_images
            if (keyword in filename if options[keyword] else keyword not in filename)
        ]

    list_of_image_urls = [
        f"{ABCLI_PUBLIC_PREFIX}/{object_name}/{image_name}"
        for image_name in tqdm(list_of_images)
    ]
    logger.info("{} image(s).".format(len(list_of_image_urls)))
    for index, image_url in enumerate(list_of_image_urls):
        logger.info(f"#{index+1} {image_url}")

    success, content, metadata = complete(
        prompt=prompt,
        detail=detail,
        list_of_image_urls=list_of_image_urls,
        verbose=verbose,
    )
    logger.info(
        "🤖 {}: {}".format(
            "success" if success else "failure",
            content,
        )
    )

    filename = objects.path_of("openai-vision.json", object_name)
    _, past_metadata = file.load_json(filename, ignore_error=True)
    past_metadata[string.timestamp()] = metadata
    if not file.save_json(filename, past_metadata, log=True):
        success = False

    return success


# https://platform.openai.com/docs/guides/vision/multiple-image-inputs
def complete(
    prompt: str,
    list_of_image_urls: List[str],
    detail: Detail = Detail.AUTO,
    max_tokens: int = 300,
    verbose: bool = False,
):
    logger.info(
        '{}.complete[{}]: "{}" @ {}'.format(
            NAME,
            detail,
            prompt,
            len(list_of_image_urls),
        )
    )

    client = OpenAI(api_key=env.OPENAI_API_KEY)
    response = client.chat.completions.create(
        model="gpt-4-vision-preview",
        messages=[
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": prompt,
                    }
                ]
                + [
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": image_urls,
                        },
                    }
                    for image_urls in list_of_image_urls
                ],
            }
        ],
        max_tokens=max_tokens,
    )

    if verbose:
        logger.info(
            "response: {}".format(
                json.dumps(
                    response.dict(),
                    indent=4,
                )
            )
        )

    if not response.choices:
        logger.info(f"{NAME}.vision.complete(): no choice.")
        return False, "", {"status": "no choice"}

    if len(response.choices) > 1:
        logger.info(
            "{} choices, picked the first, and ignored the rest.".format(
                len(response.choices)
            )
        )

    choice = response.choices[0]
    logger.info(f"{NAME}.complete(): finish_reason: {choice.finish_reason}.")

    return (
        choice.finish_reason in [None, "stop"],
        choice.message.content,
        {
            "prompt": prompt,
            "list_of_image_urls": list_of_image_urls,
            "response": response.dict(),
            "finish_reason": choice.finish_reason,
        },
    )



from typing import List
from openai_commands.logger import logger


class structured_prompt:
    def __init__(
        self,
        inputs: List[str],
        requirements: List[str],
        returns: List[str],
    ):
        self.inputs = inputs
        self.returns = returns
        self.requirements = requirements

    def create(
        self,
        function_name: str,
    ) -> str:
        output = "\n".join(
            ["Write a python function named {} that".format(function_name)]
            + (
                [
                    " inputs {} and".format(" and\n".join(self.inputs)),
                ]
                if self.inputs
                else []
            )
            + (
                [" {} and".format(" and\n".join(self.requirements))]
                if self.requirements
                else []
            )
            + [
                " returns {}.".format(" and\n".join(self.returns)),
            ]
        )

        while "  " in output:
            output = output.replace("  ", " ")

        return output



from typing import List, Dict, Any, Tuple
import random

from blue_options.host import is_jupyter
from blue_objects import file, objects

from openai_commands.completion.api import complete_prompt
from openai_commands.logger import logger


class ai_function:
    def __init__(
        self,
        function_name: str = "",
        verbose=None,
    ):
        self.language = "unknown"
        self.verbose = is_jupyter() if verbose is None else verbose

        self.function_name = (
            function_name
            if function_name
            else "{}_{}".format(
                self.__class__.__name__,
                random.randint(10000000, 99999999),
            )
        )

        self.prompt: List[str] = []
        self.code: List[str] = []

        self.auto_save = False

    def generate(
        self,
        prompt,
    ) -> Tuple[bool, Dict[str, Any]]:
        self.prompt = prompt

        logger.info("prompt: {}".format(prompt))

        success, self.code, metadata = complete_prompt(
            self.prompt,
            verbose=self.verbose,
        )

        if success:
            logger.info("code: {}".format(self.code))

        return success, metadata

    def save(self, filename=""):
        if not filename:
            filename = objects.path_of(f"{self.__class__.__name__}_code.json")

        _, content = file.load_json(filename, ignore_error=True)

        content[self.function_name] = self.to_json()

        file.save_json(filename, content)

        logger.info("-> {}".format(filename))

    def to_json(self):
        return {
            "code": self.code,
            "function_name": self.function_name,
            "prompt": self.prompt,
        }



import os
import pandas as pd
from typing import List
from flask import render_template

from blueness import module
from blue_options import string
from blue_objects import file

from openai_commands import ICON, NAME, VERSION
from openai_commands import env
from openai_commands.gpt.chat import interact_with_openai
from openai_commands.logger import logger

NAME = module.name(__file__, NAME)

examples_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../../assets/VisuaLyze")
)


class VisuaLyzeOrder:
    def __init__(
        self,
        example_name: str = "",
        prompt: List[str] = [],
        description: List[str] = [],
        data_filename: str = "",
        load_data: bool = True,
        complete: bool = False,
        log: bool = True,
        model_name: str = env.OPENAI_GPT_DEFAULT_MODEL,
    ):
        self.prompt: List[str] = prompt
        self.description: List[str] = description
        self.data_filename: str = data_filename
        self.df = pd.DataFrame()

        self.source_code: List[str] = []

        self.valid: bool = True

        if example_name:
            if not self.load_example(
                name=example_name,
                load_data=False,
                log=log,
            ):
                self.valid = False

        if load_data and self.valid:
            if not self.load_data(log=log):
                self.valid = False

        if complete and self.valid:
            if not self.complete(model_name=model_name):
                self.valid = False

        if log:
            logger.info(self.one_liner())

    def complete(
        self,
        model_name: str = env.OPENAI_GPT_DEFAULT_MODEL,
        output_path: str = "",
    ) -> bool:
        success, self.source_code = interact_with_openai(
            prompt=self.full_prompt,
            output_path=output_path,
            model_name=model_name,
        )
        if not success:
            self.valid = False
        return success

    @property
    def full_prompt(self):
        return f"{self.prompt}. {self.description}"

    def load_data(
        self,
        log: bool = True,
    ) -> bool:
        success, self.df = file.load_dataframe(
            filename=self.data_filename,
            log=log,
        )
        if not success:
            self.valid = False
        return success

    def load_example(
        self,
        name: str = "onlinefoods",
        load_data: bool = True,
        log: bool = True,
    ) -> bool:
        success, examples = file.load_yaml(os.path.join(examples_path, "examples.yaml"))
        if not success:
            return success

        path = os.path.join(examples_path, name)

        try:
            example = [example for example in examples if example["name"] == name][0]

            success, self.prompt = file.load_text(
                os.path.join(path, "prompt.txt"),
                log=log,
            )
            assert success

            success, self.description = file.load_text(
                os.path.join(path, "description.txt"),
                log=log,
            )
            assert success

            self.data_filename = os.path.join(path, "data.csv")

        except Exception as e:
            logger.error(f"failed to load example {name}: {e}.")
            return False

        return self.load_data(log=log) if load_data else True

    def one_liner(self):
        return "{} {}: {} {}: {} - {} row(s) of {}: {}".format(
            "📜 valid" if self.valid else "⚠️ invalid",
            self.__class__.__name__,
            " ".join(self.prompt),
            " ".join(self.description),
            self.data_filename,
            len(self.df),
            ", ".join(self.df.columns),
            " ".join(self.source_code),
        )

    def render(self, status: str = ""):
        return render_template(
            "index.html",
            title=f"{NAME}.{VERSION}",
            h1=f"{ICON} {NAME}.{VERSION}",
            prompt="\n".join(self.prompt),
            description="\n".join(self.description),
            data_filename=self.data_filename,
            timestamp=string.pretty_date(),
            github_url="https://github.com/kamangir/openai-commands/tree/main/openai_commands/VisuaLyze",
            dataframe_html=self.df.head().to_html(),
            status="{}{}".format(
                (
                    "<pre><code>{}</code></pre>".format("\n".join(self.source_code))
                    if self.source_code
                    else ""
                ),
                f", {status}" if status else "",
            ),
        )



from openai_commands import NAME, VERSION, DESCRIPTION, REPO_NAME
from blueness.pypi import setup

setup(
    filename=__file__,
    repo_name=REPO_NAME,
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    packages=[
        NAME,
        f"{NAME}.completion",
        f"{NAME}.completion.functions",
        f"{NAME}.completion.prompts",
        f"{NAME}.DALLE",
        f"{NAME}.gpt",
        f"{NAME}.images",
        f"{NAME}.literature_review",
        f"{NAME}.vision",
        f"{NAME}.VisuaLyze",
    ],
    include_package_data=True,
    package_data={
        NAME: [
            "config.env",
            "sample.env",
            ".abcli/**/*.sh",
        ],
    },
)
